/**
 * @file   calcmix_heu.c
 * 
 * <EN>
 * @brief  A sample plugin for calculating Gaussians
 *
 * This sample uses Julius libraries.
 * </EN>
 * 
 * <JA>
 * @brief  ʬ۷׻ץ饰
 *
 * julius Υ饤֥Ѥޤ
 * </JA>
 * 
 * @author Nakamura Hitoshi
 * @date   2009.7.24
 * 
 * $Revision: 1.0 $
 * 
 */
/**
 * Required for a file
 *   - get_plugin_info()
 *
 * Optional for a file
 *   - initialize()
 * 
 */
/**
 * Gaussian mixture calculation plugin:
 * 
 * Required:
 *   - calcmix_get_optname()
 *   - calcmix()
 *   - calcmix_init()
 *   - calcmix_free()
 * 
 */

/***************************************************************************/

/* we refer to Julius libsent header */
#include <sent/stddefs.h>
#include <sent/hmm_calc.h>
#include <julius/juliuslib.h>	// Recog¤Υ
//#include "plugin_defs.h"
#include <glib.h>		// GHashTable

#define PLUGIN_TITLE "Gaussian calculation plugin for Julius.(ADD_MASK_TO_HEU)"
#define GPRUNE_OPT "add_mask_to_heu"

static LOGPROB *eachgconst;
static GHashTable *ht;

static LOGPROB calcmix_compute_g_base(HMMWork *wrk, HTK_HMM_Dens *binfo, void *eachg);
static LOGPROB calcmix_compute_g_safe(HMMWork *wrk, HTK_HMM_Dens *binfo, LOGPROB thres, void *eachg);
static LOGPROB calcmix_compute_g_heu_updating(HMMWork *wrk, HTK_HMM_Dens *binfo, void *eachg);
static LOGPROB calcmix_compute_g_heu_pruning(HMMWork *wrk, HTK_HMM_Dens *binfo, LOGPROB thres, void *eachg);
static void calcmix_init_backmax(HMMWork *wrk);
static void calcmix_make_backmax(HMMWork *wrk);

/*	פʤΤǺ 2009/09/09 H.Nakamura
// ϥåơ֥볫ؿ
static void free_hash_table_entry(gpointer key, gpointer value, gpointer user_data)
{
  g_free(key);
  g_free(value);
}
*/

/** 
 * <EN>
 * @brief  Initialization at loading time (optional)
 * 
 * If defined, this will be called just before this plugin is loaded to Julius.
 * if this returns -1, the whole functions in this file will not be loaded.
 *
 * This function is OPTIONAL.
 * </EN>
 * <JA>
 * @brief  ɤ߹߻νǤա
 *
 * ưJulius Υץ饰ɤ߹ݤ˺ǽ˸ƤФ롥
 * -1 ֤ȡΥץ饰Τɤ߹ޤʤʤ롥
 * ¹ԲǽΥå˻Ȥ롥
 *
 * </JA>
 * 
 * 
 * @return 0 on success, -1 on failure.
 * 
 */
int
initialize()
{
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  int ret = 0;
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** initialize() : [out]return=%d\n", ret);
#endif
  return ret;
#else
  return 0;
#endif
}

/** 
 * <EN>
 * @brief  Get information of this plugin (required)
 *
 * This function should return informations of this plugin file.
 * The required info will be specified by opcode:
 *  - 0: return description string of this file into buf
 *
 * This will be called just after Julius find this file and after
 * initialize().
 *
 * @param opcode [in] requested operation code
 * @param buf [out] buffer to store the return string
 * @param buflen [in] maximum length of buf
 *
 * @return 0 on success, -1 on failure.  On failure, Julius will ignore this
 * plugin.
 * 
 * </EN>
 * <JA>
 * @brief  ץ饰ɬܡ
 *
 * Υץ饰˴ؤ֤Ϳ줿 opcode ˤäư롥
 *  - 0 ξ硤Υץ饰ե̾ΤͿ줿Хåե˳Ǽ
 *
 * δؿϡJulius Υץ饰ɤ߹ľ˸ƤФ롥
 * 
 * @param opcode [in] ׵ư ( 0 Τ߼)
 * @param buf [out] ͤǼХåե
 * @param buflen [in] buf κĹ
 * 
 * @return 顼 -1,  0 ֤顼Ȥ -1 ֤硤
 * Υץ饰Τɤ߹ޤʤ
 * </JA>
 * 
 */
int
get_plugin_info(int opcode, char *buf, int buflen)
{
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** get_plugin_info() : [in]opcode=%d\n", opcode);
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** get_plugin_info() : [in]buflen=%d\n", buflen);
#endif
#endif
  switch(opcode) {
  case 0:
    /* plugin description string */
    strncpy(buf, PLUGIN_TITLE, buflen);
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** get_plugin_info() : [out]buf=%s\n", buf);
#endif
#endif
    break;
  }

  return 0;
}

/************************************************************************/
/************************************************************************/

/** 
 * <EN>
 * @brief  Return option string to select at option. (required)
 *
 * This function should return option string which should be specified
 * as an argument "-gprune" option, to be used on Julius.  The returning
 * string should not be the same with any existing value.
 *
 * This function will be called several times at option parsing at startup.
 *
 * @param buf [out] buffer to store the return string
 * @param buflen [in] maximum length of buf
 * </EN>
 * <JA>
 * @brief  ׻ˡѥץΤʸ֤ɬܡ
 *
 * Julius ǵư "-gprune ֤" ȻꤹȤΥץ饰
 * Ѥ롥δؿǤϡ嵭 "-gprune" Ϳ٤ʸ
 * Ǽ֤֤ʸϡƥ "-gprune" ץˤǤˤ
 * Τ䡤¾Υץ饰󤬻ѤƤΤƱǤʤȡʤ⤷Ʊ
 * ä祷ƥ¦ͥ褵
 *
 * δؿϡưΥץϻ˲٤ƤФ롥
 *
 * @param buf [out] ͤǼ֤Хåե
 * @param buflen [in] buf κĹ
 * </JA>
 * 
 */
void
calcmix_get_optname(char *buf, int buflen)
{
  strncpy(buf, GPRUNE_OPT, buflen);
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_get_optname() : [in]buflen=%d\n", buflen);
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_get_optname() : [out]*buf=%s\n", buf);
#endif
#endif
}

pid_t calcmix_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/**
 * <EN>
 * @brief  A basic implementaion of computing Gaussians
 *
 * This function should compute output probabilities for each
 * Gaussians.  after this function returns, Julius will do addlog to
 * get the final output log probability.
 *
 * The input vector to be computed is located at wrk->OP_vec[], at a
 * length of wrk->OP_veclen.  Gaussians are given by g[], at a number
 * of num.  last_id and lnum is for internal use for pruning, just ignore
 * them.
 *
 * The scores for each Gaussians computed in this function should be
 * stored in OP_calced_score[], with their corresponding Gaussian ids
 * to OP_calced_id.  The total number of calculated mixtures shuold
 * also stored in OP_calced_num.
 * 
 * @param wrk [i/o] HMM computation work area to store data
 * @param g [in] set of Gaussian densities to compute the output probability.
 * @param num [in] length of above
 * @param last_id [in] ID list of N-best mixture in previous input frame,
 * or NULL if not exist
 * @param lnum [in] length of last_id
 * </EN>
 * <JA>
 * @brief  ʬ۷׻ؿ
 *
 * δؿǤϡͿ줿ʣΥʬۤФϥ٥ȥ
 * ϳΨ롥δؿԤΤϡʣΥʬۤ줾
 * ϳΨλФȳǼΤߤǤꡤʬۤȤƤνŤ߷׻ addlog
 * Ϥδؿ֤äȤ Julius ¦ǹԤ롥
 *
 * ϥ٥ȥ wrk->OP_vec[] ˳ǼƤꡤĹ wrk->OP_veclen
 * Ǥ롥ʬ g[] ȤʣϤ졤ο gnum Ǥ롥
 * 
 * ʤlast_id  lnum ϤΥʬ۽ g[] ˤľϥ
 * 졼Ƿ׻줿Τ id äƤ롥Julius ѤʤΤǡ
 * ȤʤƤ⺹٤ʤ
 *
 * ƥʬۤФϥ٥ȥпϳΨϡΥʬۤ
 * ID (0 Ϥޤź)  wrk->OP_calced_id ˡͤ
 * wrk->OP_calced_score ˳Ǽ뤳ȡޤºݤ˷׻줿
 * ʬۤο wrk->OP_calced_num ˳Ǽ뤳ȡ
 * ʤ Gaussian pruning ꤷǤ
 *
 * ʲϡpruning ԤʤñʽϳΨ׻ΤǤ롥
 * ʬۤгѶʬꤷƤ롥ʤ Julius Ǥɤ߹߻
 * HTK ǤȤ gconst ͤϤ餫׻롥Τᡤ׻
 *  dens->gconst Τ褦ѤǤ롥
 * 
 * @param wrk [i/o] HMM׻ѥꥢ
 * @param g [in] ϳΨ׻륬ʬۤ
 * @param gnum [in] @a g Υʬۤο
 * @param last_id [in] ľϥե졼Ǿ̤äʬۤIDꥹȡ
 * ޤ NULL
 * @param lnum [in] @a last_id Ĺ
 * </JA>
 */
void
calcmix(HMMWork *wrk, HTK_HMM_Dens **g, int gnum, int *last_id, int lnum)
{
  int i, j, num = 0;
  LOGPROB score, thres;
  void *eachg_array1;

  // ϥå夫value
  void *eachg_array2 = g_hash_table_lookup(ht, g);
  if (eachg_array2 == NULL){
    // ghash̤Ͽ
    printf("calcmix err ! hash̤Ͽ\n");
    return;
  }
#ifdef EACHGCONST_TEST_HASH
          printf("key=0x%x,val=0x%x\n",g, eachg_array2);
#endif

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_dbgprint_pass != pid){
    int i = 0;
    int i_tmp_buf = 0;
    float f_tmp_buf = 0;
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]wrk=0x%x(address)\n", wrk);
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]gnum=%d\n", gnum);
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]last_id=0x%x(address)\n", last_id);
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]lnum=%d\n", lnum);
#endif
#ifdef ADD_MASK_TO_HEU_DEBUG_DATA_PRINT
    for (i = 0; i < wrk->OP_veclen; i++){
      memcpy(&i_tmp_buf, &wrk->OP_vec[i], sizeof(int));
      printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]vec[%d]=%x\n", i, i_tmp_buf);
    }
    for (i = 0; i < wrk->OP_veclen; i++){
      memcpy(&f_tmp_buf, &wrk->OP_vec[i+wrk->OP_veclen], sizeof(float));
      printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix() : [in]mask[%d]=%f\n", i, f_tmp_buf);
    }
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_dbgprint_pass = pid;
#endif
  }
#endif

  if (last_id != NULL) {	/* compute them first to form thresholds */
    /* 1. clear backmax */
    calcmix_init_backmax(wrk);
    /* 2. calculate first $OP_gprune_num with setting max for each dimension */
    for (j=0; j<lnum; j++) {
      i = last_id[j];
      eachg_array1 = eachg_array2 + (i * wrk->OP_veclen * sizeof(float));
#ifdef EACHGCONST_TEST_HASH
          printf("val->0x%x\n", eachg_array1);
#endif
      score = calcmix_compute_g_heu_updating(wrk, g[i], eachg_array1);
      num = cache_push(wrk, i, score, num);
      wrk->mixcalced[i] = TRUE;      /* mark them as calculated */
    }
    /* 3. set backmax for each dimension */
    calcmix_make_backmax(wrk);
    /* 4. calculate the rest with pruning*/
    thres = wrk->OP_calced_score[num-1];
    for (i = 0; i < gnum; i++) {
      /* skip calced ones in 1. */
      if (wrk->mixcalced[i]) {
        wrk->mixcalced[i] = FALSE;
        continue;
      }
      eachg_array1 = eachg_array2 + (i * wrk->OP_veclen * sizeof(float));
#ifdef EACHGCONST_TEST_HASH
          printf("val->0x%x\n", eachg_array1);
#endif
      /* compute with safe pruning */
      score = calcmix_compute_g_heu_pruning(wrk, g[i], thres, eachg_array1);
      if (score > LOG_ZERO) {
        num = cache_push(wrk, i, score, num);
        thres = wrk->OP_calced_score[num-1];
      }
    }
  } else {	/* in case the last_id not available */
    /* at the first 0 frame */
    /* calculate with safe pruning */
    thres = LOG_ZERO;
    for (i = 0; i < gnum; i++) {
      eachg_array1 = eachg_array2 + (i * wrk->OP_veclen * sizeof(float));
#ifdef EACHGCONST_TEST_HASH
          printf("val->0x%x\n", eachg_array1);
#endif
      if (num < wrk->OP_gprune_num) {
        score = calcmix_compute_g_base(wrk, g[i], eachg_array1);
      } else {
        score = calcmix_compute_g_safe(wrk, g[i], thres, eachg_array1);
        if (score <= thres) continue;
      }
      num = cache_push(wrk, i, score, num);
      thres = wrk->OP_calced_score[num-1];
    }
  }
  wrk->OP_calced_num = num;
}

/**
 * <EN>
 * Free work area.
 * You should free all allocated at clacmix_init().
 * 
 * @param wrk [i/o] HMM computation work area
 * </EN>
 * <JA>
 * calcmix_init() ǳݤ줿ꥢ롥
 * 
 * @param wrk [i/o] HMM ׻ѥꥢ
 * </JA>
 * 
 */
void
calcmix_free(HMMWork *wrk)
{
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_free()\n");
#endif
#endif

  free(wrk->OP_calced_score);
  free(wrk->OP_calced_id);
  free(wrk->mixcalced);
  free(wrk->backmax);

  // ϥå
  //g_hash_table_foreach(ht, free_hash_table_entry, NULL);	// פʤΤǺ 2009/09/09 H.Nakamura
  if(ht) {
    g_hash_table_destroy(ht);

    free(eachgconst);

    ht = NULL;
  }
}
  
/**
 * <EN>
 * @brief  Initialize and setup work area for Gaussian computation.
 *
 * You should set value for OP_calced_maxnum, and allocate OP_calced_score
 * and OP_calced_id.  Remaining the content below is safe.
 *
 * This will be called once on instance initialization at startup.
 * 
 * @param wrk [i/o] HMM computation work area
 * 
 * @return TRUE on success, FALSE on failure.
 * </EN>
 * <JA>
 * @brief  ׻ѤΥꥢݤ롥
 *
 * ʬ۷׻ѤΥꥢݤ롥ˤǤ˽񤤤Ƥʬϡ
 * Τޤ Julius ǤѤƤΤǡʤȡ
 *
 * δؿϺǽ˲ٷ׻󥹥󥹤Ȥ˸ƤӽФ롥
 * 
 * @param wrk [i/o] HMM ׻ѥꥢ
 * 
 * @return  TRUEԻ FALSE ֤
 * </JA>
 */
boolean
calcmix_init(HMMWork *wrk)
{
  int i;
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  boolean ret = TRUE;
#endif
  /* maximum Gaussian set size = maximum mixture size * nstream */
  wrk->OP_calced_maxnum = wrk->OP_hmminfo->maxmixturenum * wrk->OP_nstream;
  wrk->OP_calced_score = (LOGPROB *)mymalloc(sizeof(LOGPROB) * wrk->OP_calced_maxnum);
  wrk->OP_calced_id = (int *)mymalloc(sizeof(int) * wrk->OP_calced_maxnum);
  wrk->mixcalced = (boolean *)mymalloc(sizeof(int) * wrk->OP_calced_maxnum);
  for(i=0;i<wrk->OP_calced_maxnum;i++) wrk->mixcalced[i] = FALSE;
  wrk->backmax_num = wrk->OP_hmminfo->opt.vec_size + 1; // vec_size+1Ƥ뤬
  wrk->backmax = (LOGPROB *)mymalloc(sizeof(LOGPROB) * wrk->backmax_num);

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_init() : [out]return=%d\n", ret);
#endif
  return ret;
#else
  return TRUE;
#endif
}

int
startup(void *data)
{
  Recog *recog = data;

  /* modified ---calculate the gconst of each vector*/
  {
    PROCESS_AM *atmp;

    HTK_HMM_State *stmp;
    GCODEBOOK *book;
    HTK_HMM_Dens **dens;
    LOGPROB *tmp;
    int i, j, k, s;
    int eachgconstsize;     // eachgconst memory size
    int totalstatenum = recog->amlist->hmminfo->totalstatenum;	// 2000
    int maxmixturenum = recog->amlist->hmminfo->maxmixturenum;	// 16
    int vecsize = recog->amlist->hmminfo->opt.vec_size;		// 27
    boolean is_tied_mixture = recog->amlist->hmminfo->is_tied_mixture;	// ǥΥ
    int mix_num = 0;
    void *key, *value;

    eachgconstsize = 0;
    for(atmp = recog->amlist, k = 1; atmp; atmp = atmp->next, k++) {
	    eachgconstsize += atmp->hmminfo->opt.vec_size * atmp->hmminfo->maxmixturenum * atmp->hmminfo->totalstatenum;
    }

#ifdef EACHGCONST_TEST_INPUT
    printf("totalstatenum=%d\nmaxmixturenum=%d\nvecsize=%d\nis_tied_mixture=%d\n", totalstatenum, maxmixturenum, vecsize, is_tied_mixture);
#endif

    // ϥå
    ht = g_hash_table_new(g_direct_hash, g_direct_equal);
#ifdef EACHGCONST_TEST_HASH
    printf("ht=0x%x\n",ht);
#endif

    // buf (julius1streamʤΤǥȥ꡼Ϲθʤ)
   tmp = eachgconst = (LOGPROB *)malloc(sizeof(LOGPROB) * eachgconstsize);
#ifdef EACHGCONST_TEST_INPUT
    printf("tmp=0x%x\n",tmp);
#endif

    // ׻
  for (atmp = recog->amlist; atmp; atmp = atmp->next) {
    for (stmp = atmp->hmminfo->ststart, k = 0; stmp; stmp = stmp->next, k++) {
      for(s = 0; s < stmp->nstream; s++) {

        if (is_tied_mixture){
          book = (GCODEBOOK *)(stmp->pdf[s]->b);
          mix_num = book->num;
          dens = book->d;
        } else {
          mix_num = stmp->pdf[s]->mix_num;
          dens = stmp->pdf[s]->b;
        }
#ifdef EACHGCONST_TEST_INPUT
        printf("[in]mix_num=%d,stream=%d\n",mix_num,stmp->nstream);
#endif

        key = dens;
        value = tmp;
        if (g_hash_table_lookup(ht, key) == NULL){  //keyʣǧ
          g_hash_table_insert(ht, key, value);
#ifdef EACHGCONST_TEST_HASH
          printf("[in]key=0x%x,val=0x%x\n",key, value);
#endif
        } else {
          // ʣξ礬ä̵뤹
          //printf("startup warning ! keyʣ key=0x%x\n", key);
        }
        
        for (j = 0; j < mix_num; j++){
#ifdef EACHGCONST_TEST_INPUT
          float eachg_add = 0.0;
#endif
#ifdef EACHGCONST_TEST_HASH
          printf("[in]b[%d]=0x%x\n",j,dens[j]);
#endif
          for (i = 0; i < vecsize; i++){
            // variance_inversed=TRUE(տˤݻ)Ǥ 
           if(dens[j] != NULL) { 
            *tmp = log(2 * PI / dens[j]->var->vec[i]);
             }
#ifdef EACHGCONST_TEST_INPUT
            printf("[%d][%d][%d]vec=%f,egc=%f\n",k,j,i,(1 / dens[j]->var->vec[i]),*tmp);
            eachg_add += *tmp;
#endif
#ifdef EACHGCONST_TEST_OUTPUT
            printf("[in]%f(0x%x)\n",*tmp,tmp);
#endif
            tmp++;
          }
#ifdef EACHGCONST_TEST_INPUT
            printf("gconst=%f,eachg_add=%f\n",dens[j]->gconst,eachg_add);
#endif
        }
        tmp += (maxmixturenum - mix_num) * vecsize;	// ɥ쥹Ĵ
      }
    }
  }
#ifdef EACHGCONST_TEST_INPUT
    // å
    {
      LOGPROB *check = eachgconst;
      for (k = 0; k < totalstatenum; k++) {
        for (j = 0; j < maxmixturenum; j++){
          for (i = 0; i < vecsize; i++){
            printf("[%d][%d][%d]vec=%f\n",k,j,i,*check++);
          }
        }
      }
    }
#endif
  }

  return 0;
}

pid_t calcmix_compute_g_heu_updating_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/**
 * @brief  Calculate probability with maximum value update.
 * 
 * Calculate probability of a Gaussian toward OP_vec,
 * while storing the maximum values of each dimension to @a backmax.
 * for future pruning.  The pruning itself is not performed here.
 * This function will be used to compute the first N Gaussians.
 * 
 * @param wrk [i/o] HMM computation work area
 * @param binfo [in] Gaussian density
 * 
 * @return the output log probability.
 */
static LOGPROB
calcmix_compute_g_heu_updating(HMMWork *wrk, HTK_HMM_Dens *binfo, void *eachg)
{
#ifdef NO_MASK_GCONST	// ޥʤgconst
  VECT tmp, x, sum = 0.0;
  VECT *mean;
  VECT *var;
  VECT *bm = wrk->backmax;
  VECT *vec = wrk->OP_vec;
  short veclen = wrk->OP_veclen;  // ϲǥΥ

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_heu_updating_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_heu_updating() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_heu_updating_dbgprint_pass = pid;
#endif
  }
#endif

  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;

  tmp = 0.0;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp = x * x * *(var++);
    sum += tmp;
    if ( *bm < tmp) *bm = tmp;
    bm++;
  }
  return((sum + binfo->gconst) * -0.5);
#else	// NO_MASK_GCONST ޥeachgconst

  VECT tmp, x, sum = 0.0, mygconst;
  VECT *mean;
  VECT *var, *myvar;
  VECT *bm = wrk->backmax;
  VECT *vec = wrk->OP_vec, *vec2;
  short veclen = wrk->OP_veclen;  // ϲǥΥ

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_heu_updating_dbgprint_pass != pid){
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_heu_updating() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_heu_updating_dbgprint_pass = pid;
#endif
  }
#endif

#ifdef EACHGCONST_TEST_HASH
          printf("eachg=0x%x\n", eachg);
#endif


  if (binfo == NULL) return(LOG_ZERO);

  vec2 = vec + veclen;	// maskϥǡȾ
  mean = binfo->mean;
  var = binfo->var->vec;
  myvar=(VECT*)eachg;
  tmp = 0.0;mygconst = 0.0;

  for (; veclen > 0; veclen--) {

#ifdef EACHGCONST_TEST_OUTPUT
    printf("%f(0x%x)\n",*myvar,myvar);
#endif

    x = *(vec++) - *(mean++);
    mygconst += *(vec2) * *(myvar++);
    tmp = *(vec2++) * (x * x * *(var++));
    sum += tmp;
    if ( *bm < tmp) *bm = tmp;
    bm++;
  }

  return((sum + mygconst) * -0.5);

#endif	// NO_MASK_GCONST ޥꡦʤȽ
}

pid_t calcmix_compute_g_heu_pruning_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/** 
 * @brief  Calculate probability with pruning.
 * 
 * Calculate probability of a Gaussian toward OP_vec,
 * performing pruning using threshold and the backmax information
 * that has been set by compute_g_heu_updating() and
 * calcmix_make_backmax().
 * 
 * @param wrk [i/o] HMM computation work area
 * @param binfo [in] Gaussian density
 * @param thres [in] threshold 
 * 
 * @return the output log probability.
 */
static LOGPROB
calcmix_compute_g_heu_pruning(HMMWork *wrk, HTK_HMM_Dens *binfo, LOGPROB thres, void *eachg)
{
#ifdef NO_MASK_GCONST	// ޥʤgconst
  VECT tmp, x;
  VECT *mean;
  VECT *var;
  VECT *bm = wrk->backmax;
  VECT *vec = wrk->OP_vec;
  short veclen = wrk->OP_veclen;  // ϲǥΥ
  LOGPROB fthres;

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_heu_pruning_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_heu_pruning() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_heu_pruning_dbgprint_pass = pid;
#endif
  }
#endif
  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
  fthres = thres * (-2.0);

  tmp = 0.0;
  bm++;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x * *(var++);
    if ( tmp + *bm > fthres) {
      return LOG_ZERO;
    }
    bm++;
  }
  return((tmp + binfo->gconst) * -0.5);

#else	// NO_MASK_GCONST ޥeachgconst

  VECT tmp, x, mygconst;
  VECT *mean;
  VECT *var, *myvar;
  VECT *bm = wrk->backmax;
  VECT *vec = wrk->OP_vec, *vec2;
  short veclen = wrk->OP_veclen;  // ϲǥΥ
  LOGPROB fthres;

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_heu_pruning_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_heu_pruning() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_heu_pruning_dbgprint_pass = pid;
#endif
  }
#endif

#ifdef EACHGCONST_TEST_HASH
          printf("eachg=0x%x\n", eachg);
#endif

  if (binfo == NULL) return(LOG_ZERO);
  vec2 = vec + veclen;	// maskϥǡȾ
  mean = binfo->mean;
  var = binfo->var->vec;
  myvar=(VECT*)eachg;
  fthres = thres * (-2.0);
  tmp = 0.0; mygconst = 0.0;

  bm++;
  for (; veclen > 0; veclen--) {
 
#ifdef EACHGCONST_TEST_OUTPUT
    printf("%f(0x%x)\n",*myvar,myvar);
#endif
    x = *(vec++) - *(mean++);
    mygconst += *(vec2) * *(myvar++);
    tmp += *(vec2++) * (x * x * *(var++));
    if ( tmp + *bm > fthres) {
      return LOG_ZERO;
    }
    bm++;
  }
  
  return((tmp + mygconst) * -0.5);
#endif	// NO_MASK_GCONST ޥꡦʤȽ
}

pid_t calcmix_init_backmax_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/** 
 * Clear backmax information.
 * 
 * @param wrk [i/o] HMM computation work area
 * 
 */
static void
calcmix_init_backmax(HMMWork *wrk)
{
  int i;
  for(i=0;i<wrk->backmax_num;i++) wrk->backmax[i] = 0;

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_init_backmax_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_init_backmax()\n");
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_init_backmax_dbgprint_pass = pid;
#endif
  }
#endif
}

pid_t calcmix_make_backmax_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/** 
 * Build backmax information for each frame, by summing up
 * current maximum values of each dimensions.
 * 
 * @param wrk [i/o] HMM computation work area
 */
/*                        |
 *  0 1 2 ... max-3 max-2 | max-1
 *
 *  a b c      x      y   | ????
 *             |
 *             v
 *  .....     x+y     y     0.0
 */
static void
calcmix_make_backmax(HMMWork *wrk)
{
  int i;

  wrk->backmax[wrk->backmax_num-1] = 0.0;
  /* backmax[len-1] = backmax[len-1];*/
  for(i=wrk->backmax_num-2;i>=0;i--) {
    wrk->backmax[i] += wrk->backmax[i+1];
  }
  /*  for(i=0;i<=len;i++) {
    printf("backmax[%d]=%f\n",i,backmax[i]);
    }*/
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_make_backmax_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
  printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_make_backmax()\n");
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_make_backmax_dbgprint_pass = pid;
#endif
  }
#endif
}

pid_t calcmix_compute_g_base_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/** 
 * Calculate probability of a Gaussian density against input
 * vector on OP_vec.
 * 
 * @param wrk [i/o] HMM computation work area
 * @param binfo [in] a Gaussian density
 * 
 * @return the output log probability.
 */
static LOGPROB
calcmix_compute_g_base(HMMWork *wrk, HTK_HMM_Dens *binfo, void *eachg)
{
#ifdef NO_MASK_GCONST	// ޥʤgconst
  VECT tmp, x;
  VECT *mean;
  VECT *var;
  VECT *vec = wrk->OP_vec;
  short veclen = wrk->OP_veclen;  // ϲǥΥ

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_base_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_base() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_base_dbgprint_pass = pid;
#endif
  }
#endif
  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
  tmp = binfo->gconst;
  for (; veclen > 0; veclen--) {
#ifdef ENABLE_MSD
    if (*vec == LZERO) {
      vec++;
      continue;
    }
#endif
    x = *(vec++) - *(mean++);
    tmp += x * x * *(var++);
  }
  return(tmp * -0.5);

#else	// NO_MASK_GCONST ޥeachgconst
  VECT tmp, x;
  VECT *mean;
  VECT *var, *myvar;
  VECT *vec = wrk->OP_vec, *vec2;
  short veclen = wrk->OP_veclen;  // ϲǥΥ
  int i;

#ifdef GCONST_COMP_TEST
  VECT comp = 0.0;
#endif
#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_base_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_base() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_base_dbgprint_pass = pid;
#endif
  }
#endif

#ifdef EACHGCONST_TEST_HASH
          printf("eachg=0x%x\n", eachg);
#endif

  if (binfo == NULL) return(LOG_ZERO);
  
  vec2 = vec + veclen;	// maskϥǡȾ
  mean = binfo->mean;
  var = binfo->var->vec;myvar=(VECT*)eachg;
  
  tmp = 0.0;
  for (i = veclen; i > 0; i--) {
#ifdef EACHGCONST_TEST_OUTPUT
    printf("%f(0x%x)\n",*myvar,myvar);
#endif

#ifdef GCONST_COMP_TEST
    comp += *myvar;
#endif

    tmp += *(vec2++) * *(myvar++);
  }
  vec2 = vec + veclen;
  
  for (; veclen > 0; veclen--) {
#ifdef ENABLE_MSD
    if (*vec == LZERO) {
      vec++;
      continue;
    }
#endif

    x = *(vec++) - *(mean++);
    tmp += *(vec2++) * (x * x * *(var++));
  }

// gconst(ޥ1ξΤ߰פ)
#ifdef GCONST_COMP_TEST
  //printf("gconst =%f\negc_vec=%f\n\n",binfo->gconst, comp);
  printf("gconst=\t%f\tegc_vec=\t%f\n",binfo->gconst, comp);
#endif

  return(tmp * -0.5);

#endif	// NO_MASK_GCONST ޥꡦʤȽ
}

pid_t calcmix_compute_g_safe_dbgprint_pass = -1;	// Ϥǽ1(ƽФPIDåȤ򤷤ʤ)
/** 
 * @brief  Calculate probability with safe pruning.
 * 
 * Calculate probability of a Gaussian toward OP_vec,
 * performing pruning using the scholar threshold.
 * 
 * @param wrk [i/o] HMM computation work area
 * @param binfo [in] Gaussian density
 * @param thres [in] threshold
 * 
 * @return the output log probability.
 */
static LOGPROB
calcmix_compute_g_safe(HMMWork *wrk, HTK_HMM_Dens *binfo, LOGPROB thres, void *eachg)
{
#ifdef NO_MASK_GCONST	// ޥʤgconst
  VECT tmp, x;
  VECT *mean;
  VECT *var;
  VECT *vec = wrk->OP_vec;
  short veclen = wrk->OP_veclen;  // ϲǥΥ
  VECT fthres = thres * (-2.0);

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_safe_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_safe() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_safe_dbgprint_pass = pid;
#endif
  }
#endif
  if (binfo == NULL) return(LOG_ZERO);
  mean = binfo->mean;
  var = binfo->var->vec;
  tmp = binfo->gconst;
  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += x * x * *(var++);
    if (tmp > fthres)  return LOG_ZERO;
  }
  return(tmp * -0.5);

#else	// NO_MASK_GCONST ޥeachgconst
  VECT tmp, x;
  VECT *mean;
  VECT *var, *myvar;
  VECT *vec = wrk->OP_vec, *vec2;
  short veclen = wrk->OP_veclen;  // ϲǥΥ
  VECT fthres = thres * (-2.0);
#ifdef GCONST_COMP_TEST
  VECT comp = 0.0;
#endif
  int i;

#ifdef ADD_MASK_TO_HEU_DEBUG_PRINT
  pid_t pid = getpid();
  if (calcmix_compute_g_safe_dbgprint_pass != pid){  // ǽ1Τ߽
#ifdef ADD_MASK_TO_HEU_DEBUG_PARAM_PRINT
    printf("***[ADD_MASK_TO_HEU_DEBUG_PRINT]*** calcmix_compute_g_safe() : [in]veclen=%d\n", veclen);
#endif
#ifdef FIRST_DBGPRINT_ONLY
    calcmix_compute_g_safe_dbgprint_pass = pid;
#endif
  }
#endif

#ifdef EACHGCONST_TEST_HASH
          printf("eachg=0x%x\n", eachg);
#endif

  if (binfo == NULL) return(LOG_ZERO);

  vec2 = vec + veclen;	// maskϥǡȾ
  mean = binfo->mean;
  var = binfo->var->vec;myvar=(VECT*)eachg;

  tmp = 0.0;
  for (i = veclen; i > 0; i--) {
#ifdef EACHGCONST_TEST_OUTPUT
    printf("%f(0x%x)\n",*myvar,myvar);
#endif

#ifdef GCONST_COMP_TEST
    comp += *myvar;
#endif

    tmp += *(vec2++) * *(myvar++);
  }
  vec2 = vec + veclen;

  for (; veclen > 0; veclen--) {
    x = *(vec++) - *(mean++);
    tmp += *(vec2++) * (x * x * *(var++));
    if (tmp > fthres)  return LOG_ZERO;
  }

// gconst(ޥ1ξΤ߰פ)
#ifdef GCONST_COMP_TEST
  //printf("gconst =%f\negc_vec=%f\n\n",binfo->gconst, comp);
  printf("gconst=\t%f\tegc_vec=\t%f\n",binfo->gconst, comp);
#endif

  return(tmp * -0.5);
#endif	// NO_MASK_GCONST ޥꡦʤȽ
}

/* end of file */
